#include <Wire.h>

#define IMU 0x28
#define IMURST A0
#define LED 7
#define write8(reg, data) Wire.beginTransmission(IMU);Wire.write(reg);Wire.write(data);Wire.endTransmission()

void IMU_waitforboot() { //wait for IMU boot, attempt to read and verify ID
  do {
    delay(10);
    Wire.beginTransmission(IMU);
    Wire.write(0x00);
    Wire.endTransmission(false);
    Wire.requestFrom(IMU, 1, true);
  } while(Wire.read() != 0xA0);
}

byte IMU_checkcalibration() { //returns 0 (not calibrated) - 3 (fully calibrated)
  Wire.beginTransmission(IMU);
  Wire.write(0x35);
  Wire.endTransmission(false);
  Wire.requestFrom(IMU, 1, true);
  byte a = Wire.read();
  byte b = a;
  Serial.print("CALIBRATION: ");
  Serial.print(a, BIN);
  Serial.print(" SYS=");
  Serial.print(a >> 6);
  a &= 0b00111111;
  Serial.print(" GYR=");
  Serial.print(a >> 4);
  a &= 0b00001111;
  Serial.print(" ACC=");
  Serial.print(a >> 2);
  a &= 0b00000011;
  Serial.print(" MAG=");
  Serial.println(a);
  return b >> 6;
}

void IMU_syscheck() {
  IMU_checkcalibration();
  byte a;
  Wire.beginTransmission(IMU);
  Wire.write(0x36);
  Wire.endTransmission(false);
  Wire.requestFrom(IMU, 5, true);
  Serial.print("ST RESULT: ");
  a = Wire.read();
  Serial.print(a, BIN);
  if (a & 0b1000) {
    Serial.print(" MCU OK");
  }
  if (a & 0b100) {
    Serial.print(" GYR OK");
  }
  if (a & 0b10) {
    Serial.print(" MAG OK");
  }
  if (a & 0b1) {
    Serial.print(" ACC OK");
  }
  Wire.read();
  Wire.read();
  Serial.println();
  Serial.print("SYSTEM STATUS: ");
  switch(Wire.read()) {
    case 0:
      Serial.println("0 - System idle");
      break;
    case 1:
      Serial.println("1 - System Error");
      break;
    case 2:
      Serial.println("2 - Initializing peripherals");
      break;
    case 3:
      Serial.println("3 - System Init ialization");
      break;
    case 4:
      Serial.println("4 - Executing selftest");
      break;
    case 5:
      Serial.println("5 - Sensor fusion algorithm running");
      break;
    case 6:
      Serial.println("6 - System running without fusion algorithm");
      break;
  }
  Serial.print("SYSTEM ERROR: ");
  switch(Wire.read()) {
    case 0:
      Serial.println("0 - No error");
      break;
    case 1:
      Serial.println("1 - Peripheral init ialization error");
      break;
    case 2:
      Serial.println("2 - System init ialization error");
      break;
    case 3:
      Serial.println("3 - Self test result failed");
      break;
    case 4:
      Serial.println("4 - Register map value out of range");
      break;
    case 5:
      Serial.println("5 - Register map address out of range");
      break;
    case 6:
      Serial.println("6 - Register map write error");
      break;
    case 7:
      Serial.println("7 - BNO low power mode not available for selected operation mode");
      break;
    case 8:
      Serial.println("8 - Accelerometer power mode not available");
      break;
    case 9:
      Serial.println("9 - Fusion algorithm configuration error");
      break;
    case 0xA:
      Serial.println("A - Sensor configuration error");
      break;
  }
  Wire.beginTransmission(IMU);
  Wire.write(0x3D);
  Wire.endTransmission(false);
  Wire.requestFrom(IMU, 2, true);
  Serial.print("OPERATION MODE: ");
  switch (Wire.read() & 0b00001111) {
    case 0x00:
      Serial.println("CONFIG");
      break;
    case 0x01:
      Serial.println("ACCONLY");
      break;
    case 0x02:
      Serial.println("MAGONLY");
      break;
    case 0x03:
      Serial.println("GYRONLY");
      break;
    case 0x04:
      Serial.println("ACCMAG");
      break;
    case 0x05:
      Serial.println("ACCGYRO");
      break;
    case 0x06:
      Serial.println("MAGGYRO");
      break;
    case 0x07:
      Serial.println("AMG");
      break;
    case 0x08:
      Serial.println("IMU");
      break;
    case 0x09:
      Serial.println("COMPASS");
      break;
    case 0x0A:
      Serial.println("M4G");
      break;
    case 0x0B:
      Serial.println("NDOF_FMC_OFF");
      break;
    case 0x0C:
      Serial.println("NDOF");
      break;
  }
  Serial.print("POWER MODE: ");
  switch (Wire.read()) {
    case 0:
      Serial.println("NORMAL");
      break;
    case 1:
      Serial.println("LOW POWER");
      break;
    case 2:
      Serial.println("SUSPEND");
      break;
  }
}

void IMU_rawdata() {
  Wire.beginTransmission(IMU);
  Wire.write(0x08);
  Wire.endTransmission(false);
  Wire.requestFrom(IMU, 18, true);
  // Accelerometer
  Serial.print("ACC X=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/100.00, 2); // m/s^2
  Serial.print(" Y=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/100.00, 2); // m/s^2
  Serial.print(" Z=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/100.00, 2); // m/s^2
  // Magnetometer
  Serial.print(" MAG X=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // mT
  Serial.print(" Y=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // mT
  Serial.print(" Z=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // mT
  // Gyroscope
  Serial.print(" GYR X=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // deg/s
  Serial.print(" Y=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // deg/s
  Serial.print(" Z=");
  Serial.println((int16_t)(Wire.read()|Wire.read()<<8 )/16.00, 2); // deg/s
}

void IMU_euler() {
  Wire.beginTransmission(IMU);
  Wire.write(0x1A);  
  Wire.endTransmission(false);
  Wire.requestFrom(IMU, 6, true);
  Serial.print("Yaw=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00);
  Serial.print(" Roll=");
  Serial.print((int16_t)(Wire.read()|Wire.read()<<8 )/16.00);
  Serial.print(" Pitch=");
  Serial.println((int16_t)(Wire.read()|Wire.read()<<8 )/16.00);
}

void setup() {
  pinMode(LED, OUTPUT);
  pinMode(IMURST, OUTPUT);
  digitalWrite(IMURST, 1);
  digitalWrite(LED, 0);
  Wire.begin();

  IMU_waitforboot();

  Serial.begin(115200);  //Setting the baudrate
  delay(5000);
  Serial.println("Start");
  digitalWrite(LED, 1);

  write8(0x3F, 0b01000000); // System trigger: Internal oscillator, reset

  IMU_waitforboot();

  write8(0x3F, 0b00000001); // System trigger: Internal oscillator, self test

  IMU_waitforboot();

  write8(0x3E, 0x00); // Power Mode: Normal: 0X00 (or B00), Low Power: 0X01 (or B01) , Suspend Mode: 0X02 (orB10)

  delay(50);

  write8(0x41, 0x21); // AXIS MAP CONFIG
  write8(0x42, 0x04); // AXIS MAP SIGN

  write8(0x3B, 0b0001000); //android orientation, fahrenheit, degrees, deg/s, m/s^2
  
  write8(0x3D, 0x0C); // Operation Mode: NDOF:0X0C (or B1100) , IMU:0x08 (or B1000) , NDOF_FMC_OFF: 0x0B (or B1011), GYROONLY 0b0011 / 0x03, AMG 0b0111 / 0x07

  delay(100);

  IMU_waitforboot();

  Serial.println("----------INIT DONE----------");
  delay(10000);
  //while(IMU_checkcalibration() < 3) {
  //  delay(500);
  //}
  Serial.println("----------CALIBRATED----------");
  IMU_syscheck();
}

void loop() {
  for(int i = 0; i < 50; i++) {
    IMU_rawdata();
    IMU_euler();
    delay(100);
  }
  IMU_syscheck();
}
